package org.scalaquery.simple

import java.sql.{SQLException, Date, Time, Timestamp}
import org.scalaquery.session.PositionedParameters

/**
 * Basic conversions for setting parameters in PositionedParameters.
 */
trait SetParameter[-T] extends ((T, PositionedParameters) => Unit)

object SetParameter {
  implicit object SetBoolean extends SetParameter[Boolean] { def apply(v: Boolean, pp: PositionedParameters) { pp.setBoolean(v) } }
  implicit object SetByte extends SetParameter[Byte] { def apply(v: Byte, pp: PositionedParameters) { pp.setByte(v) } }
  implicit object SetDate extends SetParameter[Date] { def apply(v: Date, pp: PositionedParameters) { pp.setDate(v) } }
  implicit object SetDouble extends SetParameter[Double] { def apply(v: Double, pp: PositionedParameters) { pp.setDouble(v) } }
  implicit object SetFloat extends SetParameter[Float] { def apply(v: Float, pp: PositionedParameters) { pp.setFloat(v) } }
  implicit object SetInt extends SetParameter[Int] { def apply(v: Int, pp: PositionedParameters) { pp.setInt(v) } }
  implicit object SetLong extends SetParameter[Long] { def apply(v: Long, pp: PositionedParameters) { pp.setLong(v) } }
  implicit object SetShort extends SetParameter[Short] { def apply(v: Short, pp: PositionedParameters) { pp.setShort(v) } }
  implicit object SetString extends SetParameter[String] { def apply(v: String, pp: PositionedParameters) { pp.setString(v) } }
  implicit object SetTime extends SetParameter[Time] { def apply(v: Time, pp: PositionedParameters) { pp.setTime(v) } }
  implicit object SetTimestamp extends SetParameter[Timestamp] { def apply(v: Timestamp, pp: PositionedParameters) { pp.setTimestamp(v) } }

  implicit object SetBooleanOption extends SetParameter[Option[Boolean]] { def apply(v: Option[Boolean], pp: PositionedParameters) { pp.setBooleanOption(v) } }
  implicit object SetByteOption extends SetParameter[Option[Byte]] { def apply(v: Option[Byte], pp: PositionedParameters) { pp.setByteOption(v) } }
  implicit object SetDateOption extends SetParameter[Option[Date]] { def apply(v: Option[Date], pp: PositionedParameters) { pp.setDateOption(v) } }
  implicit object SetDoubleOption extends SetParameter[Option[Double]] { def apply(v: Option[Double], pp: PositionedParameters) { pp.setDoubleOption(v) } }
  implicit object SetFloatOption extends SetParameter[Option[Float]] { def apply(v: Option[Float], pp: PositionedParameters) { pp.setFloatOption(v) } }
  implicit object SetIntOption extends SetParameter[Option[Int]] { def apply(v: Option[Int], pp: PositionedParameters) { pp.setIntOption(v) } }
  implicit object SetLongOption extends SetParameter[Option[Long]] { def apply(v: Option[Long], pp: PositionedParameters) { pp.setLongOption(v) } }
  implicit object SetShortOption extends SetParameter[Option[Short]] { def apply(v: Option[Short], pp: PositionedParameters) { pp.setShortOption(v) } }
  implicit object SetStringOption extends SetParameter[Option[String]] { def apply(v: Option[String], pp: PositionedParameters) { pp.setStringOption(v) } }
  implicit object SetTimeOption extends SetParameter[Option[Time]] { def apply(v: Option[Time], pp: PositionedParameters) { pp.setTimeOption(v) } }
  implicit object SetTimestampOption extends SetParameter[Option[Timestamp]] { def apply(v: Option[Timestamp], pp: PositionedParameters) { pp.setTimestampOption(v) } }

  implicit def createSetTuple2[T1, T2](implicit c1: SetParameter[T1], c2: SetParameter[T2]): SetParameter[(T1, T2)] = new SetParameter[(T1, T2)] {
    def apply(param: (T1, T2), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
    }
  }
  implicit def createSetTuple3[T1, T2, T3](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3]): SetParameter[(T1, T2, T3)] = new SetParameter[(T1, T2, T3)] {
    def apply(param: (T1, T2, T3), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
    }
  }
  implicit def createSetTuple4[T1, T2, T3, T4](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4]): SetParameter[(T1, T2, T3, T4)] = new SetParameter[(T1, T2, T3, T4)] {
    def apply(param: (T1, T2, T3, T4), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
    }
  }
  implicit def createSetTuple5[T1, T2, T3, T4, T5](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5]): SetParameter[(T1, T2, T3, T4, T5)] = new SetParameter[(T1, T2, T3, T4, T5)] {
    def apply(param: (T1, T2, T3, T4, T5), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
    }
  }
  implicit def createSetTuple6[T1, T2, T3, T4, T5, T6](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6]): SetParameter[(T1, T2, T3, T4, T5, T6)] = new SetParameter[(T1, T2, T3, T4, T5, T6)] {
    def apply(param: (T1, T2, T3, T4, T5, T6), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
    }
  }
  implicit def createSetTuple7[T1, T2, T3, T4, T5, T6, T7](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7]): SetParameter[(T1, T2, T3, T4, T5, T6, T7)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
    }
  }
  implicit def createSetTuple8[T1, T2, T3, T4, T5, T6, T7, T8](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
    }
  }
  implicit def createSetTuple9[T1, T2, T3, T4, T5, T6, T7, T8, T9](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
    }
  }
  implicit def createSetTuple10[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
    }
  }
  implicit def createSetTuple11[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
    }
  }
  implicit def createSetTuple12[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
    }
  }
  implicit def createSetTuple13[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
    }
  }
  implicit def createSetTuple14[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
    }
  }
  implicit def createSetTuple15[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
    }
  }
  implicit def createSetTuple16[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
    }
  }
  implicit def createSetTuple17[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
    }
  }
  implicit def createSetTuple18[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17], c18: SetParameter[T18]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
      c18(param._18, pp)
    }
  }
  implicit def createSetTuple19[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17], c18: SetParameter[T18], c19: SetParameter[T19]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
      c18(param._18, pp)
      c19(param._19, pp)
    }
  }
  implicit def createSetTuple20[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17], c18: SetParameter[T18], c19: SetParameter[T19], c20: SetParameter[T20]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
      c18(param._18, pp)
      c19(param._19, pp)
      c20(param._20, pp)
    }
  }
  implicit def createSetTuple21[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17], c18: SetParameter[T18], c19: SetParameter[T19], c20: SetParameter[T20], c21: SetParameter[T21]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
      c18(param._18, pp)
      c19(param._19, pp)
      c20(param._20, pp)
      c21(param._21, pp)
    }
  }
  implicit def createSetTuple22[T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22](implicit c1: SetParameter[T1], c2: SetParameter[T2], c3: SetParameter[T3], c4: SetParameter[T4], c5: SetParameter[T5], c6: SetParameter[T6], c7: SetParameter[T7], c8: SetParameter[T8], c9: SetParameter[T9], c10: SetParameter[T10], c11: SetParameter[T11], c12: SetParameter[T12], c13: SetParameter[T13], c14: SetParameter[T14], c15: SetParameter[T15], c16: SetParameter[T16], c17: SetParameter[T17], c18: SetParameter[T18], c19: SetParameter[T19], c20: SetParameter[T20], c21: SetParameter[T21], c22: SetParameter[T22]): SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22)] = new SetParameter[(T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22)] {
    def apply(param: (T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22), pp: PositionedParameters) {
      c1(param._1, pp)
      c2(param._2, pp)
      c3(param._3, pp)
      c4(param._4, pp)
      c5(param._5, pp)
      c6(param._6, pp)
      c7(param._7, pp)
      c8(param._8, pp)
      c9(param._9, pp)
      c10(param._10, pp)
      c11(param._11, pp)
      c12(param._12, pp)
      c13(param._13, pp)
      c14(param._14, pp)
      c15(param._15, pp)
      c16(param._16, pp)
      c17(param._17, pp)
      c18(param._18, pp)
      c19(param._19, pp)
      c20(param._20, pp)
      c21(param._21, pp)
      c22(param._22, pp)
    }
  }

  object SetSimpleProduct extends SetParameter[Product] {
    def apply(prod: Product, pp: PositionedParameters): Unit =
      for(i <- 0 until prod.productArity) prod.productElement(i) match {
        case v: Boolean => pp.setBoolean(v)
        case v: Byte => pp.setByte(v)
        case v: Date => pp.setDate(v)
        case v: Double => pp.setDouble(v)
        case v: Float => pp.setFloat(v)
        case v: Int => pp.setInt(v)
        case v: Long => pp.setLong(v)
        case v: Short => pp.setShort(v)
        case v: String => pp.setString(v)
        case v: Time => pp.setTime(v)
        case v: Timestamp => pp.setTimestamp(v)
        case v => throw new SQLException("SetProduct doesn't know how to handle parameter "+i+" ("+v+")")
      }
  }

  implicit object SetUnit extends SetParameter[Unit] { def apply(none: Unit, pp: PositionedParameters) = () }

  def apply[T](implicit f: (T, PositionedParameters) => Unit) = new SetParameter[T] {
    def apply(v: T, pp: PositionedParameters) = f(v, pp)
  }
}
